import hashlib
from cryptography.hazmat.primitives.asymmetric import rsa, padding
from cryptography.hazmat.primitives import serialization, hashes
from cryptography import x509
from cryptography.x509.oid import NameOID
import datetime
#Backend: OpenSSL
#Serialization is converting a data object (e.g., Python objects, Tensorflow models) into a format that allows us to store or transmit the data
#Privacy-Enhanced Mail (PEM) is a de facto file format for storing and sending cryptographic keys, certificates
#Optimal Asymmetric Encryption Padding (OAEP) is a padding scheme often used together with RSA encryption
#A mask generation function (MGF) is a cryptographic primitive similar to a cryptographic hash function except that while a hash function's output has a fixed size, a MGF supports output of a variable length.
#Mask generation functions were first proposed as part of the specification for padding in the RSA-OAEP algorithm.
#MGF1 is a mask generation function defined in the Public Key Cryptography Standard #1
#Probabilistic Signature Scheme (PSS) is a cryptographic signature scheme
#Valid paddings for signatures are PSS and PKCS1v15. PSS is the recommended choice for any new protocols or applications

def hash_message(message):
    try:
        digest = hashlib.sha256(message.encode()).hexdigest()
        print(f"Hashed Message: {digest}")
    except Exception as e:
        print(f"Hashing failed: {e}")

message = "Hello from cryptography"
#hash_message(message)      

private_key = None
ciphertext = ""
def gen_rsa_keypair(keylength, private_key_pass="default"):
    global private_key
    private_key = rsa.generate_private_key(public_exponent=65537, key_size=keylength)
    encrypted_pem_private_key = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            #encryption_algorithm=serialization.NoEncryption()
            encryption_algorithm=serialization.BestAvailableEncryption(private_key_pass.encode("utf-8"))
        )
    pem_public_key = private_key.public_key().public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )

    with open("private-rsa.pem", "wb") as private_key_file:
            private_key_file.write(encrypted_pem_private_key)

    with open("public-rsa.pem", "wb") as public_key_file:
        public_key_file.write(pem_public_key)

    print("RSA key pair generation successful!")

#gen_rsa_keypair(2048, "Cdac@123")
#gen_rsa_keypair(1024)
    
def encrypt_message(message):
    try:
        global ciphertext
        with open("public-rsa.pem", "rb") as key_file:
            public_key = serialization.load_pem_public_key(
            key_file.read()
        )
        #public_key = private_key.public_key()
        ciphertext = public_key.encrypt(
            message.encode("utf-8"),
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None
            )
        )
        print(f"Encrypted Message: {ciphertext.hex()}")
    except Exception as e:
        print(f"Encryption failed: {e}")    

#encrypt_message("Hi! I am Anoop Pandey!")

def decrypt_message():
    try:
        global ciphertext
        with open("private-rsa.pem", "rb") as key_file:
            private_key = serialization.load_pem_private_key(
            key_file.read(),
            password="Cdac@123".encode(),
        )
        plaintext = private_key.decrypt(
            ciphertext,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None
            )
        )
        print(f"Decrypted Message: {plaintext.decode('utf-8')}")
    except Exception as e:
        print(f"Decryption failed: {e}")

#decrypt_message()
signature = None        
def digital_sign_message(message, passphrase):
    try:
        global signature
        with open("private-rsa.pem", "rb") as key_file:
            private_key = serialization.load_pem_private_key(
            key_file.read(),
            password=passphrase.encode(),
        )
        signature = private_key.sign(
            message.encode(),
            padding.PSS(
                mgf=padding.MGF1(hashes.SHA256()),
                salt_length=padding.PSS.MAX_LENGTH
            ),
            hashes.SHA256()
        )
        print(f"Digital Signature: {signature.hex()}")
    except Exception as e:
        print(f"Error signing message: {e}")        

#digital_sign_message("Hi! I am Anoop", "Cdac@123")

def verify_digital_signature():
    try:
        global signature
        message = "Hi! I am Anoop"
        with open("public-rsa.pem", "rb") as key_file:
            public_key = serialization.load_pem_public_key(
            key_file.read()
        )
        public_key.verify(
            signature,
            message.encode(),
            padding.PSS(
                mgf=padding.MGF1(hashes.SHA256()),
                salt_length=padding.PSS.MAX_LENGTH
            ),
            hashes.SHA256()
        )
        print("Digital Signature is valid.")
    except Exception as e:
       print(f"Verification failed: {type(e).__name__}")

#verify_digital_signature()

def create_digital_signature_certificate(passphrase):
    try:
        country_name = input("Enter Country Name (e.g., IN): ")
        state_name = input("Enter State or Province Name: ")
        locality_name = input("Enter Locality Name: ")
        organization_name = input("Enter Organization Name: ")
        common_name = input("Enter Common Name: ")
        subject = x509.Name([
            x509.NameAttribute(NameOID.COUNTRY_NAME, country_name),
            x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, state_name),
            x509.NameAttribute(NameOID.LOCALITY_NAME, locality_name),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, organization_name),
            x509.NameAttribute(NameOID.COMMON_NAME, common_name),
        ])
        with open("private-rsa.pem", "rb") as key_file:
            private_key = serialization.load_pem_private_key(
            key_file.read(),
            password=passphrase.encode(),
        )
        cert = x509.CertificateBuilder().subject_name(
            subject
        ).issuer_name(
            subject
        ).public_key(
            private_key.public_key()
        ).serial_number(
            x509.random_serial_number()
        ).not_valid_before(
            datetime.datetime.utcnow()
        ).not_valid_after(
            datetime.datetime.utcnow() + datetime.timedelta(days=365)
        ).sign(
            private_key, hashes.SHA256()
        )
        filename = common_name+"_certificate.pem"
        with open(filename, "wb") as cert_file:
            cert_file.write(cert.public_bytes(serialization.Encoding.PEM))

        print(f"Digital Signature Certificate for {common_name} created and saved to {filename}.")
    except Exception as e:
        print(f"Error creating digital signature certificate: {e}")

create_digital_signature_certificate("Cdac@123")